package test;

import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import libFLAC.JFLAC__BitReaderReadCallback;

/** test_libFLAC - Unit tester for libFLAC
 * Copyright (C) 2000-2009  Josh Coalson
 * Copyright (C) 2011-2022  Xiph.Org Foundation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
public final class Jbitreader implements JFLAC__BitReaderReadCallback {
	private static final int FLAC__BYTES_PER_WORD = 4;// JFLAC__BitReader.FLAC__BYTES_PER_WORD
	private static final int FLAC__BITS_PER_WORD = 32;// JFLAC__BitReader.FLAC__BITS_PER_WORD
	private static final boolean WORDS_BIGENDIAN = false;
/*
 * WATCHOUT!  Since FLAC__BitReader is a private structure, we use a copy of
 * the definition here to get at the internals.  Make sure this is kept up
 * to date with what is in ../libFLAC/bitreader.c
 */
/* #if (ENABLE_64_BIT_WORDS == 0)

typedef FLAC__uint32 brword;
#define FLAC__BYTES_PER_WORD 4
#define FLAC__BITS_PER_WORD 32

#else

typedef FLAC__uint64 brword;
#define FLAC__BYTES_PER_WORD 8
#define FLAC__BITS_PER_WORD 64

#endif */

/* struct FLAC__BitReader {
	// any partially-consumed word at the head will stay right-justified as bits are consumed from the left
	// any incomplete word at the tail will be left-justified, and bytes from the read callback are added on the right
	int[] buffer;
	int capacity; // in words
	int words; // # of completed words in buffer
	int bytes; // # of bytes in incomplete word at buffer[words]
	int consumed_words; // #words ...
	int consumed_bits; // ... + (#bits of head word) already consumed from the front of buffer
	int read_crc16; // the running frame CRC
	int crc16_offset; // the number of words in the current buffer that should not be CRC'd
	int crc16_align; // the number of bits in the current consumed word that should not be CRC'd
	JFLAC__BitReaderReadCallback read_callback;
	// Object client_data;
//}; */

	private static Object newJFLAC__BitReader() {
		try {
			final Class<?> ref = Class.forName("libFLAC.JFLAC__BitReader");
			final Constructor<?> constructor = ref.getDeclaredConstructor();
			constructor.setAccessible( true );
			return constructor.newInstance();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final InstantiationException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		} catch( final ClassNotFoundException e ) {
			e.printStackTrace();
		}
		return null;
	}

	private static boolean callMethodBool(final Object br, final String methodName) {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName );
			method.setAccessible( true );
			return ((Boolean) method.invoke( br )).booleanValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return false;
	}

	private static int callMethodInt(final Object br, final String methodName) {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName );
			method.setAccessible( true );
			return ((Integer) method.invoke( br )).intValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return 0;
	}

	private static void callMethodVoidNoThrow(final Object br, final String methodName, final Class<?>[] paramTypes, final Object... args) {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			method.invoke( br, args );
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
	}

	private static void callMethodVoid(final Object br, final String methodName, final Class<?>[] paramTypes, final Object... args) throws IOException {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			method.invoke( br, args );
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
	}

	private static boolean callMethodBool(final Object br, final String methodName, final Class<?>[] paramTypes, final Object... args) {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			return ((Boolean) method.invoke( br, args )).booleanValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return false;
	}

	private static int callMethodInt(final Object br, final String methodName, final Class<?>[] paramTypes, final Object... args) throws IOException {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			return ((Integer) method.invoke( br, args )).intValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return 0;
	}

	private static long callMethodLong(final Object br, final String methodName, final Class<?>[] paramTypes, final Object... args) throws IOException {
		try {
			final Method method = br.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			return ((Long) method.invoke( br, args )).longValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return 0;
	}

	private final byte[] callback_data;

	Jbitreader(final byte[] data) {
		callback_data = data;
	}

	@SuppressWarnings("boxing")
	static boolean test_bitreader()
	{
		Object /* JFLAC__BitReader */ br;
		boolean ok = true;
		int i;
		int words, bits; /* what we think br->consumed_words and br->consumed_bits should be */

		int crc;
		final char[] expected_crcs = { 0x5e4c, 0x7f6b, 0x2272, 0x42bf };
		final byte[] data = new byte[32];

		int val_uint32;
		long  val_uint64;

		for( i = 0; i < 32; i++ ) {
			data[i] = (byte)(i * 8 + 7);
		}

		final JFLAC__BitReaderReadCallback read_callback = new Jbitreader( data );

		System.out.printf("\n+++ libFLAC unit test: bitreader\n\n");

		/*
		 * test new -> delete
		 */
		System.out.printf("testing new... ");
		br = newJFLAC__BitReader();// new JFLAC__BitReader();

		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		br = null;// br.FLAC__bitreader_delete();
		System.out.printf("OK\n");

		/*
		 * test new -> init -> delete
		 */
		System.out.printf("testing new... ");
		br = newJFLAC__BitReader();// new JFLAC__BitReader();

		System.out.printf("OK\n");

		System.out.printf("testing init... ");
		if( ! callMethodBool( br, "FLAC__bitreader_init", new Class<?>[]{JFLAC__BitReaderReadCallback.class}, read_callback ) ) {// if( ! br.FLAC__bitreader_init( read_callback ) ) {
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		br = null;// FLAC__bitreader_delete(br);
		System.out.printf("OK\n");

		/*
		 * test new -> init -> clear -> delete
		 */
		System.out.printf("testing new... ");
		br = newJFLAC__BitReader();// new JFLAC__BitReader();

		System.out.printf("OK\n");

		System.out.printf("testing init... ");
		if( ! callMethodBool( br, "FLAC__bitreader_init", new Class<?>[]{JFLAC__BitReaderReadCallback.class}, read_callback ) ) {// if( ! br.FLAC__bitreader_init( read_callback ) ) {
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing clear... ");
		if( ! callMethodBool( br, "FLAC__bitreader_clear" ) ) { // if( ! br.FLAC__bitreader_clear() ) {
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		br = null;// FLAC__bitreader_delete(br);
		System.out.printf("OK\n");

		/*
		 * test normal usage
		 */
		System.out.printf("testing new... ");
		br = newJFLAC__BitReader();// new JFLAC__BitReader();

		System.out.printf("OK\n");

		System.out.printf("testing init... ");
		if( ! callMethodBool( br, "FLAC__bitreader_init", new Class<?>[]{JFLAC__BitReaderReadCallback.class}, read_callback ) ) {// if( ! br.FLAC__bitreader_init( read_callback ) ) {
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing clear... ");
		if( ! callMethodBool( br, "FLAC__bitreader_clear" ) ) { // if( ! br.FLAC__bitreader_clear() ) {
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		words = bits = 0;

		System.out.printf("capacity = %d\n", Jtests.getIntegerValue( br, "capacity") );

		System.out.printf("testing raw reads... ");
		/*
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 10 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 4 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 32 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 4 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint64 = br.FLAC__bitreader_read_raw_uint64( 64 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 12); } catch(final IOException e) { ok = false; }
		*/

		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 10 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 4 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 32 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 4 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint64 = callMethodLong( br, "FLAC__bitreader_read_raw_uint64", new Class<?>[]{int.class}, 64 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 12 ); } catch(final IOException e) { ok = false; }

		if( ! ok ) {
			System.out.printf("FAILED\n");
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		/* we read 152 bits (=19 bytes) from the bitreader */
		words = 152 / FLAC__BITS_PER_WORD;
		bits = 152 - words * FLAC__BITS_PER_WORD;

		int val = Jtests.getIntegerValue( br, "consumed_words");
		if( val != words ) {
			System.out.printf("FAILED word count %d != %d\n", val, words );
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( br, "consumed_bits");
		if( val != bits ) {
			System.out.printf("FAILED bit count %d != %d\n", val, bits);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != expected_crcs[0] ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0x%04x\n", crc, expected_crcs[0]);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");
		callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );

		System.out.printf("testing CRC reset... ");
		callMethodBool( br, "FLAC__bitreader_clear" );// br.FLAC__bitreader_clear();
		callMethodVoidNoThrow( br, "FLAC__bitreader_reset_read_crc16", new Class<?>[]{int.class}, 0xFFFF );// br.FLAC__bitreader_reset_read_crc16( 0xFFFF );
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != 0xFFFF ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0xFFFF\n", crc);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		callMethodVoidNoThrow( br, "FLAC__bitreader_reset_read_crc16", new Class<?>[]{int.class}, 0 );// br.FLAC__bitreader_reset_read_crc16( 0 );
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != 0 ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0x0000\n", crc);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		// try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 16 ); } catch(final IOException e) {}
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 16 ); } catch(final IOException e) {}
		callMethodVoidNoThrow( br, "FLAC__bitreader_reset_read_crc16", new Class<?>[]{int.class}, 0 );// br.FLAC__bitreader_reset_read_crc16( 0 );
		// try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 32 ); } catch(final IOException e) {}
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 32 ); } catch(final IOException e) {}
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != expected_crcs[1] ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0x%04x\n", crc, expected_crcs[1]);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing unaligned < 32 bit reads... ");
		callMethodBool( br, "FLAC__bitreader_clear" );// br.FLAC__bitreader_clear();
		// try { br.FLAC__bitreader_skip_bits_no_crc( 8 ); } catch(final IOException e) {}
		try { callMethodVoid( br, "FLAC__bitreader_skip_bits_no_crc", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) {}
		callMethodVoidNoThrow( br, "FLAC__bitreader_reset_read_crc16", new Class<?>[]{int.class}, 0 );// br.FLAC__bitreader_reset_read_crc16( 0 );
		/*
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 8 ); } catch(final IOException e) { ok = false; }
		*/
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) { ok = false; }

		if( ! ok ) {
			System.out.printf("FAILED\n");
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != expected_crcs[2] ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0x%04x\n", crc, expected_crcs[2]);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");
		callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );

		System.out.printf("testing unaligned < 64 bit reads... ");
		callMethodBool( br, "FLAC__bitreader_clear" );// br.FLAC__bitreader_clear();
		// try { br.FLAC__bitreader_skip_bits_no_crc( 8 ); } catch(final IOException e) {}
		try { callMethodVoid( br, "FLAC__bitreader_skip_bits_no_crc", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) {}
		callMethodVoidNoThrow( br, "FLAC__bitreader_reset_read_crc16", new Class<?>[]{int.class}, 0 );// br.FLAC__bitreader_reset_read_crc16( 0 );
		/*
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = br.FLAC__bitreader_read_raw_uint32( 32 ); } catch(final IOException e) { ok = false; }
		*/
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 1 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 2 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 5 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 8 ); } catch(final IOException e) { ok = false; }
		try { val_uint32 = callMethodInt( br, "FLAC__bitreader_read_raw_uint32", new Class<?>[]{int.class}, 32 ); } catch(final IOException e) { ok = false; }
		if( ! ok ) {
			System.out.printf("FAILED\n");
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		crc = callMethodInt( br, "FLAC__bitreader_get_read_crc16");// crc = br.FLAC__bitreader_get_read_crc16();
		if( crc != expected_crcs[3] ) {
			System.out.printf("FAILED reported CRC 0x%04x does not match expected 0x%04x\n", crc, expected_crcs[3]);
			callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");
		callMethodVoidNoThrow( br, "FLAC__bitreader_dump", new Class<?>[]{PrintStream.class}, System.out );// br.FLAC__bitreader_dump( System.out );

		System.out.printf("testing free... ");
		//br.FLAC__bitreader_clear();
		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		br = null;// br.FLAC__bitreader_delete();
		System.out.printf("OK\n");

		System.out.printf("\nPASSED!\n");
		return true;
	}

	/*----------------------------------------------------------------------------*/

	@Override
	public final int bit_read_callback(final byte[] buffer, int bytes) throws IOException {
		if( bytes > 32 ) {
			bytes = 32;
		}

		System.arraycopy( callback_data, 0, buffer, 0, bytes );

		return bytes;
	}
}