package test;

import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/** test_libFLAC - Unit tester for libFLAC
 * Copyright (C) 2000-2009  Josh Coalson
 * Copyright (C) 2011-2022  Xiph.Org Foundation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
public final class Jbitwriter {
	private static final int FLAC__BYTES_PER_WORD = 4;// JFLAC__BitWriter.FLAC__BYTES_PER_WORD
	private static final int FLAC__BITS_PER_WORD = 32;// JFLAC__BitWriter.FLAC__BITS_PER_WORD
	private static final boolean WORDS_BIGENDIAN = true;// java uses big endian to store bytes into int arrays
/*
 * WATCHOUT!  Since FLAC__BitWriter is a private structure, we use a copy of
 * the definition here to get at the internals.  Make sure this is kept up
 * to date with what is in ../libFLAC/bitwriter.c
 */
/* #if (ENABLE_64_BIT_WORDS == 0)

typedef FLAC__uint32 bwword;
#define FLAC__BYTES_PER_WORD 4
#define FLAC__BITS_PER_WORD 32
#define PRI_BWWORD "08x"

#else

typedef FLAC__uint64 bwword;
#define FLAC__BYTES_PER_WORD 8
#define FLAC__BITS_PER_WORD 64
#define PRI_BWWORD "016" PRIx64

#endif */

/* struct FLAC__BitWriter {
	int[] buffer;
	int accum; // accumulator; bits are right-justified; when full, accum is appended to buffer
	int capacity; // capacity of buffer in words
	int words; // # of complete words in buffer
	int bits; // # of used bits in accum
}; */

// #define WORDS_TO_BITS(words) ((words) * FLAC__BITS_PER_WORD)
// #define TOTAL_BITS(bw) (WORDS_TO_BITS((bw)->words) + (bw)->bits)

	private static int WORDS_TO_BITS(final int words) {
		return words * FLAC__BITS_PER_WORD;
	}

	private static int TOTAL_BITS(final Object bw) {
		final int bits = Jtests.getIntegerValue( bw, "bits" );
		final int words = Jtests.getIntegerValue( bw, "words" );
		return WORDS_TO_BITS( words ) + bits;
	}

	private static Object newFLAC__BitWriter() {
		try {
			final Class<?> ref = Class.forName("libFLAC.JFLAC__BitWriter");
			final Constructor<?> constructor = ref.getDeclaredConstructor();
			constructor.setAccessible( true );
			return constructor.newInstance();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final InstantiationException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		} catch( final ClassNotFoundException e ) {
			e.printStackTrace();
		}
		return null;
	}

	private static boolean callMethodBool(final Object bw, final String methodName) {
		try {
			final Method method = bw.getClass().getDeclaredMethod( methodName );
			method.setAccessible( true );
			return ((Boolean) method.invoke( bw )).booleanValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return false;
	}

	private static void callMethodVoid(final Object bw, final String methodName) {
		try {
			final Method method = bw.getClass().getDeclaredMethod( methodName );
			method.setAccessible( true );
			method.invoke( bw );
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
	}

	private static void callMethodVoid(final Object bw, final String methodName, final Class<?>[] paramTypes, final Object... args) {
		try {
			final Method method = bw.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			method.invoke( bw, args );
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
	}

	private static boolean callMethodBool(final Object bw, final String methodName, final Class<?>[] paramTypes, final Object... args) {
		try {
			final Method method = bw.getClass().getDeclaredMethod( methodName, paramTypes );
			method.setAccessible( true );
			return ((Boolean) method.invoke( bw, args )).booleanValue();
		} catch( final IllegalArgumentException e ) {
			e.printStackTrace();
		} catch( final IllegalAccessException e ) {
			e.printStackTrace();
		} catch( final InvocationTargetException e ) {
			e.printStackTrace();
		} catch( final SecurityException e ) {
			e.printStackTrace();
		} catch( final NoSuchMethodException e ) {
			e.printStackTrace();
		}
		return false;
	}

	@SuppressWarnings("boxing")
	static boolean test_bitwriter()
	{
		Object /* JFLAC__BitWriter */ bw;
		boolean ok;
		int i, j;
		int test_pattern1[];
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		test_pattern1 = new int[]{ 0xaaf0aabe, 0xaaaaaaa8, 0x300aaaaa, 0xaaadeadb, 0x00eeface };
} else {
		test_pattern1 = new int[]{ 0xbeaaf0aa, 0xa8aaaaaa, 0xaaaa0a30, 0xdbeaadaa, 0x00eeface };
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
/*if( WORDS_BIGENDIAN ) {
		test_pattern1 = new long[]{ (0xaaf0aabeaaaaaaa8L), (0x300aaaaaaaadeadbL), (0x0000000000eefaceL) };
} else {
		test_pattern1 = new long[]{ (0xa8aaaaaabeaaf0aaL), (0xdbeaadaaaaaa0a30L), (0x0000000000eefaceL) };
}*/
} else {
		// #error FLAC__BYTES_PER_WORD is neither 4 nor 8 -- not implemented
		System.out.printf("FLAC__BYTES_PER_WORD is neither 4 nor 8 -- not implemented\n");
		return false;
}
		int words, bits; /* what we think bw->words and bw->bits should be */

		System.out.printf("\n+++ libFLAC unit test: bitwriter\n\n");

		/*
		 * test new -> init -> delete
		 */
		System.out.printf("testing new... ");
		bw = newFLAC__BitWriter();//new JFLAC__BitWriter();

		System.out.printf("OK\n");

		System.out.printf("testing init... ");
		if( ! callMethodBool( bw, "FLAC__bitwriter_init") ) {// bw.FLAC__bitwriter_init()
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		bw = null;
		System.out.printf("OK\n");

		/*
		 * test normal usage
		 */
		System.out.printf("testing new... ");
		bw = newFLAC__BitWriter();//new JFLAC__BitWriter();

		System.out.printf("OK\n");

		System.out.printf("testing init... ");
		if( ! callMethodBool( bw, "FLAC__bitwriter_init") ) {// bw.FLAC__bitwriter_init()
			System.out.printf("FAILED, returned false\n");
			return false;
		}
		System.out.printf("OK\n");

		System.out.printf("testing clear... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear");// bw.FLAC__bitwriter_clear();
		System.out.printf("OK\n");

		words = bits = 0;

		int val = Jtests.getIntegerValue( bw, "capacity" );
		System.out.printf("capacity = %d\n", val );

		System.out.printf("testing zeroes, raw_uint32*... ");
		ok =
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x1, 1 ) &&// bw.FLAC__bitwriter_write_raw_uint32( 0x1, 1 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x1, 2 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0x1, 2 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xa, 5 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0xa, 5 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xf0, 8 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0xf0, 8 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x2aa, 10 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0x2aa, 10 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xf, 4 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0xf, 4 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xaaaaaaaa, 32 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0xaaaaaaaa, 32 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_zeroes", new Class<?>[]{int.class}, 4 ) &&//bw.FLAC__bitwriter_write_zeroes( 4 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x3, 2 ) &&//bw.FLAC__bitwriter_write_raw_uint32( 0x3, 2 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_zeroes", new Class<?>[]{int.class}, 8 ) &&//bw.FLAC__bitwriter_write_zeroes( 8 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint64", new Class<?>[]{long.class, int.class}, (0xaaaaaaaadeadbeefL), 64 ) &&//bw.FLAC__bitwriter_write_raw_uint64( (0xaaaaaaaadeadbeefL), 64 ) &&
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xace, 12 )//bw.FLAC__bitwriter_write_raw_uint32( 0xace, 12 )
		;
		if( ! ok ) {
			System.out.printf("FAILED\n");
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		/* we wrote 152 bits (=19 bytes) to the bitwriter */
		words = 152 / FLAC__BITS_PER_WORD;
		bits = 152 - words*FLAC__BITS_PER_WORD;

		val = Jtests.getIntegerValue( bw, "words" );
		if( val != words ) {
			System.out.printf("FAILED word count %d != %d\n", val, words);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( bw, "bits" );
		if( val != bits ) {
			System.out.printf("FAILED bit count %d != %d\n", val, bits);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		int[] buffer = Jtests.getIntBuffer( bw, "buffer" );
		if( Jtests.memcmp( buffer, 0, test_pattern1, 0, words ) != 0 ) {
			System.out.printf("FAILED pattern match (buffer)\n");
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( bw, "accum" );
		if( (val & 0x00ffffff) != test_pattern1[words] ) {
			System.out.printf("FAILED pattern match (bw.accum=%d != %d\n", val & 0x00ffffff, test_pattern1[words]);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");
		callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );

		System.out.printf("testing raw_uint32 some more... ");
		ok = callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x3d, 6 );// bw.FLAC__bitwriter_write_raw_uint32( 0x3d, 6 );
		if( ! ok ) {
			System.out.printf("FAILED\n");
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		bits += 6;
		test_pattern1[words] <<= 6;
		test_pattern1[words] |= 0x3d;
		val = Jtests.getIntegerValue( bw, "words" );
		if( val != words ) {
			System.out.printf("FAILED word count %d != %d\n", val, words);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( bw, "bits" );
		if( val != bits ) {
			System.out.printf("FAILED bit count %d != %d\n", val, bits);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		if( Jtests.memcmp( buffer, 0, test_pattern1, 0, words ) != 0 ) {
			System.out.printf("FAILED pattern match (buffer)\n");
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( bw, "accum" );
		if( (val & 0x3fffffff) != test_pattern1[words] ) {
			System.out.printf("FAILED pattern match (bw.accum=%d != %d\n", val & 0x3fffffff, test_pattern1[words]);
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		System.out.printf("OK\n");
		callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );

		System.out.printf("testing utf8_uint32(0x00000000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x00000000 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x00000000 );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 8 && (val & 0xff) == 0;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x0000007F)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x0000007F );// bw.FLAC__bitwriter_write_utf8_uint32( 0x0000007F );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 8 && (val & 0xff) == 0x7F;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x00000080)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x00000080 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x00000080 );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 16 && (val & 0xffff) == 0xC280;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x000007FF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x000007FF );// bw.FLAC__bitwriter_write_utf8_uint32( 0x000007FF );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 16 && (val & 0xffff) == 0xDFBF;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x00000800)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x00000800 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x00000800 );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 24 && (val & 0xffffff) == 0xE0A080;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x0000FFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x0000FFFF );// bw.FLAC__bitwriter_write_utf8_uint32( 0x0000FFFF );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 24 && (val & 0xffffff) == 0xEFBFBF;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x00010000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x00010000 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x00010000 );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xF0908080;
} else {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0x808090F0;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 32 && (val & 0xffffffff) == 0xF0908080;
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x001FFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x001FFFFF );// bw.FLAC__bitwriter_write_utf8_uint32( 0x001FFFFF );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xF7BFBFBF;
} else {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xBFBFBFF7;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 32 && (val & 0xffffffff) == 0xF7BFBFBF;
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x00200000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x00200000 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x00200000 );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xF8888080 && (val & 0xff) == 0x80;
} else {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0x808088F8 && (val & 0xff) == 0x80;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 40 && (val & (0xffffffffffL)) == (0xF888808080L);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x03FFFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x03FFFFFF );// bw.FLAC__bitwriter_write_utf8_uint32( 0x03FFFFFF );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xFBBFBFBF && (val & 0xff) == 0xBF;
} else {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xBFBFBFFB && (val & 0xff) == 0xBF;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 40 && (val & (0xffffffffffL)) == (0xFBBFBFBFBFL);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x04000000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x04000000 );// bw.FLAC__bitwriter_write_utf8_uint32( 0x04000000 );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xFC848080 && (val & 0xffff) == 0x8080;
} else {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0x808084FC && (val & 0xffff) == 0x8080;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 48 && (val & (0xffffffffffffL)) == (0xFC8480808080L);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint32(0x7FFFFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint32", new Class<?>[]{int.class}, 0x7FFFFFFF );// bw.FLAC__bitwriter_write_utf8_uint32( 0x7FFFFFFF );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xFDBFBFBF && (val & 0xffff) == 0xBFBF;
} else {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xBFBFBFFD && (val & 0xffff) == 0xBFBF;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 48 && (val & (0xffffffffffffL)) == (0xFDBFBFBFBFBFL);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000000000000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000000000000L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000000000000L) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 8 && (val & 0xff) == 0;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x000000000000007F)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x000000000000007FL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x000000000000007FL) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 8 && (val & 0xff) == 0x7F;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000000000080)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000000000080L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000000000080L) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 16 && (val & 0xffff) == 0xC280;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x00000000000007FF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x00000000000007FFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x00000000000007FFL) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 16 && (val & 0xffff) == 0xDFBF;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000000000800)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000000000800L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000000000800L) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 24 && (val & 0xffffff) == 0xE0A080;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x000000000000FFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x000000000000FFFFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x000000000000FFFFL) );
		val = Jtests.getIntegerValue( bw, "accum" );
		ok = TOTAL_BITS(bw) == 24 && (val & 0xffffff) == 0xEFBFBF;
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000000010000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000000010000L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000000010000L) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xF0908080;
} else {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0x808090F0;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 32 && (val & 0xffffffff) == 0xF0908080;
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x00000000001FFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x00000000001FFFFFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x00000000001FFFFFL) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xF7BFBFBF;
} else {
		ok = TOTAL_BITS(bw) == 32 && buffer[0] == 0xBFBFBFF7;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 32 && (val & 0xffffffff) == 0xF7BFBFBF;
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000000200000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000000200000L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000000200000L) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xF8888080 && (val & 0xff) == 0x80;
} else {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0x808088F8 && (val & 0xff) == 0x80;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 40 && (val & (0xffffffffffL)) == (0xF888808080L);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000003FFFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000003FFFFFFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000003FFFFFFL) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xFBBFBFBF && (val & 0xff) == 0xBF;
} else {
		ok = TOTAL_BITS(bw) == 40 && buffer[0] == 0xBFBFBFFB && (val & 0xff) == 0xBF;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 40 && (val & (0xffffffffffL)) == (0xFBBFBFBFBFL);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000004000000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000004000000L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000004000000L) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xFC848080 && (val & 0xffff) == 0x8080;
} else {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0x808084FC && (val & 0xffff) == 0x8080;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 48 && (val & (0xffffffffffffL)) == (0xFC8480808080L);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x000000007FFFFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x000000007FFFFFFFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x000000007FFFFFFFL) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xFDBFBFBF && (val & 0xffff) == 0xBFBF;
} else {
		ok = TOTAL_BITS(bw) == 48 && buffer[0] == 0xBFBFBFFD && (val & 0xffff) == 0xBFBF;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 48 && (val & (0xffffffffffffL)) == (0xFDBFBFBFBFBFL);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000080000000)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000080000000L );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000080000000L) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 56 && buffer[0] == 0xFE828080 && (val & 0xffffff) == 0x808080;
} else {
		ok = TOTAL_BITS(bw) == 56 && buffer[0] == 0x808082FE && (val & 0xffffff) == 0x808080;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 56 && (val & (0xffffffffffffffL)) == (0xFE828080808080L);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing utf8_uint64(0x0000000FFFFFFFFF)... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_utf8_uint64", new Class<?>[]{long.class}, 0x0000000FFFFFFFFFL );// bw.FLAC__bitwriter_write_utf8_uint64( (0x0000000FFFFFFFFFL) );
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == 56 && buffer[0] == 0xFEBFBFBF && (val & 0xffffff) == 0xBFBFBF;
} else {
		ok = TOTAL_BITS(bw) == 56 && buffer[0] == 0xBFBFBFFE && (val & 0xffffff) == 0xBFBFBF;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
		ok = TOTAL_BITS(bw) == 56 && (val & (0xffffffffffffffL)) == (0xFEBFBFBFBFBFBFL);
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}

		System.out.printf("testing grow... ");
		callMethodVoid( bw, "FLAC__bitwriter_clear" );// bw.FLAC__bitwriter_clear();
		callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0x5, 4 );// bw.FLAC__bitwriter_write_raw_uint32( 0x5, 4 );
		j = Jtests.getIntegerValue( bw, "capacity" );
		for( i = 0; i < j; i++ ) {
			callMethodBool( bw, "FLAC__bitwriter_write_raw_uint32", new Class<?>[]{int.class, int.class}, 0xaaaaaaaa, 32 );// bw.FLAC__bitwriter_write_raw_uint32( 0xaaaaaaaa, 32 );
		}
		buffer = Jtests.getIntBuffer( bw, "buffer" );
		val = Jtests.getIntegerValue( bw, "accum" );
if( FLAC__BYTES_PER_WORD == 4 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == i*32+4 && buffer[0] == 0x5aaaaaaa && (val & 0xf) == 0xa;
} else {
		ok = TOTAL_BITS(bw) == i*32+4 && buffer[0] == 0xaaaaaa5a && (val & 0xf) == 0xa;
}
} else if( FLAC__BYTES_PER_WORD == 8 ) {
if( WORDS_BIGENDIAN ) {
		ok = TOTAL_BITS(bw) == i*32+4 && buffer[0] == (0x5aaaaaaaaaaaaaaaL) && (val & 0xf) == 0xa;
} else {
		ok = TOTAL_BITS(bw) == i*32+4 && buffer[0] == (0xaaaaaaaaaaaaaa5aL) && (val & 0xf) == 0xa;
}
}
		System.out.printf("%s\n", ok?"OK":"FAILED");
		if( ! ok ) {
			callMethodVoid( bw, "FLAC__bitwriter_dump", new Class<?>[]{PrintStream.class}, System.out );// bw.FLAC__bitwriter_dump( System.out );
			return false;
		}
		val = Jtests.getIntegerValue( bw, "capacity" );
		System.out.printf("capacity = %d\n", val );

		System.out.printf("testing free... ");
		callMethodVoid( bw, "FLAC__bitwriter_free" );// bw.FLAC__bitwriter_free();
		System.out.printf("OK\n");

		System.out.printf("testing delete... ");
		bw = null;// bw.FLAC__bitwriter_delete();
		System.out.printf("OK\n");

		System.out.printf("\nPASSED!\n");
		return true;
	}
}